/*
 * swamigui.h - Main public header file for Swami GUI
 *
 * Swami
 * Copyright (C) 1999-2014 Element Green <element@elementsofsound.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License only.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA or point your web browser to http://www.gnu.org.
 */
#ifndef __SWAMI_GUI_H__
#define __SWAMI_GUI_H__

#include <gtk/gtk.h>

#include <swamigui/SwamiguiBar.h>
#include <swamigui/SwamiguiBarPtr.h>
#include <swamigui/SwamiguiCanvasMod.h>
#include <swamigui/SwamiguiControl.h>
#include <swamigui/SwamiguiControlAdj.h>
#include <swamigui/SwamiguiControlMidiKey.h>
#include <swamigui/SwamiguiDnd.h>
#include <swamigui/SwamiguiItemMenu.h>
#include <swamigui/SwamiguiKnob.h>
#include <swamigui/SwamiguiLoopFinder.h>
#include <swamigui/SwamiguiMenu.h>
#include <swamigui/SwamiguiModEdit.h>
#include <swamigui/SwamiguiMultiSave.h>
#include <swamigui/SwamiguiNoteSelector.h>
#include <swamigui/SwamiguiPanel.h>
#include <swamigui/SwamiguiPanelSelector.h>
#include <swamigui/SwamiguiPanelSF2Gen.h>
#include <swamigui/SwamiguiPanelSF2GenEnv.h>
#include <swamigui/SwamiguiPanelSF2GenMisc.h>
#include <swamigui/SwamiguiPaste.h>
#include <swamigui/SwamiguiPiano.h>
#include <swamigui/SwamiguiPref.h>
#include <swamigui/SwamiguiProp.h>
#include <swamigui/SwamiguiPythonView.h>
#include <swamigui/SwamiguiRoot.h>
#include <swamigui/SwamiguiSampleCanvas.h>
#include <swamigui/SwamiguiSampleEditor.h>
#include <swamigui/SwamiguiSpectrumCanvas.h>
#include <swamigui/SwamiguiSpinScale.h>
#include <swamigui/SwamiguiSplits.h>
#include <swamigui/SwamiguiStatusbar.h>
#include <swamigui/SwamiguiTree.h>
#include <swamigui/SwamiguiTreeStore.h>
#include <swamigui/SwamiguiTreeStorePatch.h>
#include <swamigui/SwamiguiTreeStoreConfig.h>
#include <swamigui/builtin_enums.h>
#include <swamigui/help.h>
#include <swamigui/icons.h>
#include <swamigui/patch_funcs.h>
#include <swamigui/splash.h>
#include <swamigui/swami_python.h>
#include <swamigui/util.h>

#endif
