LIBRARY
EXPORTS

swami_deinit

_swami_object_init
;_swami_param_init

_swami_plugin_initialize
swami_plugin_unload_all

_swami_ret_g_log

;_swami_set_patch_prop_origin_event
;_swami_state_types_init
;_swami_util_init

;group_parent_class DATA
;item_parent_class DATA
;patch_add_parent_class DATA
;patch_change_parent_class DATA
;patch_remove_parent_class DATA
root_signals DATA
;state_parent_class DATA
;swami_add_patch_prop_callback
;swami_clear_item_prop_change_cache
swami_control_conn_flags_get_type
swami_control_conn_priority_get_type
swami_control_connect
;swami_control_debug DATA
swami_control_connect_item_prop
swami_control_connect_transform
swami_control_disconnect
swami_control_disconnect_all
swami_control_disconnect_unref
swami_control_do_event_expiration
swami_control_event_active_ref
swami_control_event_active_unref
swami_control_event_duplicate
swami_control_event_free
swami_control_event_get_type
swami_control_event_new
swami_control_event_ref
swami_control_event_set_origin
swami_control_event_stamp
swami_control_event_unref
swami_control_flags_get_type
swami_control_func_assign_funcs
swami_control_func_get_type
swami_control_func_new
swami_control_get_connections
swami_control_get_flags
swami_control_get_type
swami_control_get_value
swami_control_get_value_native
swami_control_hub_get_type
swami_control_hub_new
swami_control_midi_get_type
swami_control_midi_new
swami_control_midi_send
swami_control_midi_set_callback
swami_control_midi_transmit
swami_control_new
swami_control_new_event

;swami_control_prop_assign_property
swami_control_prop_get_type
swami_control_prop_new
swami_control_queue_add_event
swami_control_queue_get_type
swami_control_queue_new
swami_control_queue_run
swami_control_queue_set_test_func

;swami_control_ref_queue
;swami_control_ref_spec

swami_control_set_event
swami_control_set_event_no_queue
swami_control_set_event_no_queue_loop
swami_control_set_flags
swami_control_set_queue
swami_control_set_spec
swami_control_set_value
swami_control_set_value_no_queue
swami_control_set_value_type

;swami_control_slave_spec

swami_control_transmit_event
swami_control_transmit_event_loop
swami_control_transmit_value
swami_control_value_alloc_value
swami_control_value_assign_value
swami_control_value_get_type
swami_control_value_new
swami_error_get_type

swami_error_quark

swami_event_item_add_copy
swami_event_item_add_free
swami_event_item_add_get_type
swami_event_item_remove_copy
swami_event_item_remove_free
swami_event_item_remove_get_type
swami_event_item_remove_new
swami_event_prop_change_copy
swami_event_prop_change_free
swami_event_prop_change_get_type
swami_event_prop_change_new
swami_find_object_property
swami_get_root
swami_init
swami_list_object_properties
swami_lock_get_atomic
swami_lock_get_type
swami_lock_set_atomic
swami_marshal_VOID__OBJECT_UINT
swami_midi_device_close
swami_midi_device_get_type
swami_midi_device_new
swami_midi_device_open

;swami_midi_device_ref_control
swami_midi_event_bank_select
swami_midi_event_control
swami_midi_event_control14
swami_midi_event_copy
swami_midi_event_free
swami_midi_event_get_type
swami_midi_event_new
swami_midi_event_note_off
swami_midi_event_note_on
swami_midi_event_nrpn
swami_midi_event_pitch_bend
swami_midi_event_rpn
swami_midi_event_set
swami_midi_event_set_bend_range
swami_midi_event_set_program
swami_midi_event_type_get_type
swami_object_clear_flags
swami_object_find_by_type
swami_object_flags_get_type
swami_object_get
swami_object_get_flags
swami_object_get_property
swami_object_get_valist
swami_object_propbag_quark DATA

;swami_object_ref_by_name
;swami_object_ref_by_type
;swami_object_remove
;swami_object_remove_recursive
;swami_object_replace

swami_object_get_origin
swami_object_set
swami_object_set_default
swami_object_set_flags
swami_object_set_property
swami_object_set_valist
swami_param_convert
swami_param_convert_new

;swami_param_get_digits

swami_param_get_limits

;swami_param_set_digits

swami_param_set_limits
swami_param_type_from_value_type

;swami_patch_add_control DATA

;swami_patch_add_control DATA
;Getter function returning swami_patch_add_control.
swami_patch_get_add_control

;swami_patch_load_ref

;swami_patch_prop_control DATA

;swami_patch_prop_title_control
;Getter function returning swami_patch_prop_title_control.
swami_patch_get_prop_title_control


;swami_patch_remove_control DATA
;Getter function returning swami_patch_remove_control.
swami_patch_get_remove_control

;swami_patch_save

swami_plugin_add_path
swami_plugin_find
swami_plugin_get_list
swami_plugin_get_type
swami_plugin_is_loaded
swami_plugin_load
swami_plugin_load_xml
swami_plugin_save_xml
swami_plugin_load_absolute
swami_plugin_load_all
swami_plugin_load_plugin

swami_prop_tree_add_value
swami_prop_tree_get_children
swami_prop_tree_get_type
swami_prop_tree_insert_before
swami_prop_tree_new

;swami_prop_tree_node_flags_get_type

swami_prop_tree_object_get_node
swami_prop_tree_prepend
swami_prop_tree_remove
swami_prop_tree_remove_recursive
swami_prop_tree_remove_value
swami_prop_tree_replace
swami_prop_tree_set_root

;swami_prop_tree_xml_save_state
swami_rank_get_type

;swami_remove_patch_prop_callback
;swami_remove_patch_prop_callback_matched

swami_root_add_object
swami_root_get_objects
swami_root_get_type
swami_root_insert_object_before
swami_root_new
swami_root_new_object
swami_root_patch_save
swami_root_prepend_object

;swami_state_begin_group
;swami_state_end_group
;swami_state_get_type
;swami_state_get_undo_depends
;swami_state_group_flags_get_type

;swami_state_group_get_type
;swami_state_item_conflict
;swami_state_item_depend
;swami_state_item_describe
;swami_state_item_flags_get_type
;swami_state_item_get_type
;swami_state_item_restore
;swami_state_item_type_get_type
;swami_state_new
;swami_state_patch_add_get_type
;swami_state_patch_change_get_type
;swami_state_patch_remove_get_type
;swami_state_record
;swami_state_record_item
;swami_state_ref_active_group
;swami_state_retract
;swami_state_set_active_group
;swami_state_undo

swami_type_get_children
swami_type_get_default
swami_type_get_rank
swami_type_set_rank

swami_util_free_value
swami_util_get_child_types
swami_util_new_value
swami_util_midi_str_to_note
swami_util_midi_note_to_str

swami_wavetbl_close

;swami_wavetbl_get_temp_item_locale

swami_wavetbl_get_control
swami_wavetbl_get_type
swami_wavetbl_load_patch

;swami_wavetbl_load_temp_item
swami_wavetbl_open

;swami_wavetbl_ref_control
;swami_wavetbl_set_gen_realtime
;swami_wavetbl_set_temp_item_locale
;swami_xml_is_doc
;swami_xml_new_doc
;swami_xml_object_get_type
;swami_xml_object_get_xml

;swami_xml_object_set_xml

;swami_xml_read_file
;swami_xml_restore_ctx_add_fixup
;swami_xml_restore_ctx_fixup
;swami_xml_restore_ctx_get_type
;swami_xml_restore_ctx_new
;swami_xml_restore_object

;swami_xml_restore_object_state
;swami_xml_restore_object_state_by_type
;swami_xml_restore_properties_by_type

;swami_xml_save_ctx_get_object_id
;swami_xml_save_ctx_get_type
;swami_xml_save_ctx_new
;swami_xml_save_object
;swami_xml_save_object_node
;swami_xml_save_object_state
;swami_xml_save_object_state_by_type
;swami_xml_save_properties_by_type
;swami_xml_state_get_type
;swami_xml_write_file

swami_loop_finder_new
swami_loop_results_get_values
swami_loop_finder_get_results
swami_loop_finder_get_type
swami_loop_finder_verify_params
swami_loop_finder_find

swami_wavetbl_get_active_item_locale
swami_object_get_by_type
swami_get_control_prop_by_name

swami_control_get_spec
swami_control_prop_connect_objects
swami_param_type_transformable
swami_get_control_prop
swami_container_get_type
swami_object_set_origin
swami_root_patch_is_loaded
swami_root_patch_load
swami_control_prop_connect_to_control
;swami_patch_add_control



wavetbl_class DATA
