/*
 * Translatable strings file generated by Glade.
 * Add this file to your project's POTFILES.in.
 * DO NOT compile it as part of your application.
 */

gchar *s = N_("Preferences");
gchar *s = N_("Restart drivers for changes to take effect");
gchar *s = N_("Type");
gchar *s = N_("Device");
gchar *s = N_("Buffer Size");
gchar *s = N_("Buffer Count");
gchar *s = N_("AUTO");
gchar *s = N_("*");
gchar *s = N_("Default");
gchar *s = N_("Audio Driver");
gchar *s = N_("Type");
gchar *s = N_("Device");
gchar *s = N_("NONE");
gchar *s = N_("*");
gchar *s = N_("Default");
gchar *s = N_("MIDI Driver");
gchar *s = N_("FluidSynth Control");
gchar *s = N_("Master Gain");
gchar *s = N_("Default");
gchar *s = N_("Disabled");
gchar *s = N_("Default *");
gchar *s = N_("Custom");
gchar *s = N_("Room Size");
gchar *s = N_("Damp");
gchar *s = N_("Width");
gchar *s = N_("Level");
gchar *s = N_("Reverb");
gchar *s = N_("Disabled");
gchar *s = N_("Default *");
gchar *s = N_("Custom");
gchar *s = N_("N");
gchar *s = N_("Level");
gchar *s = N_("Frequency");
gchar *s = N_("Depth");
gchar *s = N_("Waveform");
gchar *s = N_("Sine");
gchar *s = N_("Triangle");
gchar *s = N_("Chorus");
gchar *s = N_("* marks options that require a driver restart to take effect");
gchar *s = N_("Close");
