LIBRARY
EXPORTS

swamigui_deinit

swamigui_control_glade_prop_connect
swamigui_spectrum_canvas_get_type
swamigui_control_adj_new
swamigui_note_selector_new
swamigui_spectrum_canvas_set_data
swamigui_panel_get_type
swamigui_register_panel_selector_type

; swamigui_root DATA
; getter function to access swamigui_root
swamigui_get_swamigui_root

swamigui_root_activate
swamigui_root_load_prefs
swamigui_root_patch_load
swamigui_root_new

swamigui_knob_get_adjustment
swamigui_knob_get_type
swamigui_control_new_for_widget
swamigui_control_new_for_widget_full
swamigui_control_prop_connect_widget

swamigui_util_glade_create
swamigui_util_glade_lookup

swamigui_register_pref_handler

swamigui_disable_plugins
swamigui_disable_python

swamigui_init