/*
 * Translatable strings file generated by Glade.
 * Add this file to your project's POTFILES.in.
 * DO NOT compile it as part of your application.
 */

gchar *s = N_("Find");
gchar *s = N_("File:");
gchar *s = N_("Type:");
gchar *s = N_("All Files");
gchar *s = N_("Preset");
gchar *s = N_("Instrument");
gchar *s = N_("Sample");
gchar *s = N_("Start from beginning");
gchar *s = N_("Preset:");
gchar *s = N_("Name:");
gchar *s = N_("Sub string");
gchar *s = N_("Starts with");
gchar *s = N_("Bank:");
gchar *s = N_("Search");
gchar *s = N_("Close");
gchar *s = N_("Cancel");
gchar *s = N_("Percussion");
gchar *s = N_("Name:");
gchar *s = N_("Bank:");
gchar *s = N_("Preset:");
gchar *s = N_("Name:");
gchar *s = N_("Name:");
gchar *s = N_("Rate:");
gchar *s = N_("Name of sample");
gchar *s = N_("MIDI note number of original pitch");
gchar *s = N_("-");
gchar *s = N_("Sample rate");
gchar *s = N_("48000");
gchar *s = N_("44100");
gchar *s = N_("22050");
gchar *s = N_("11025");
gchar *s = N_("Tuning:");
gchar *s = N_("Fine tuning in cents");
gchar *s = N_("cents");
gchar *s = N_("Note:");
gchar *s = N_("New Sample");
gchar *s = N_("Name:");
gchar *s = N_("Import loop information");
gchar *s = N_("Raw Sample Parameters");
gchar *s = N_("Rate:");
gchar *s = N_("Custom:");
gchar *s = N_("16 bit");
gchar *s = N_("8 bit");
gchar *s = N_("Mono");
gchar *s = N_("Stereo");
gchar *s = N_("44100");
gchar *s = N_("22050");
gchar *s = N_("11025");
gchar *s = N_("Custom");
gchar *s = N_("Width:");
gchar *s = N_("Channels:");
gchar *s = N_("Endian:");
gchar *s = N_("Little");
gchar *s = N_("Big");
gchar *s = N_("Sign:");
gchar *s = N_("Signed");
gchar *s = N_("Unsigned");
gchar *s = N_("OK");
gchar *s = N_("Cancel");
gchar *s = N_("Swami Tips");
gchar *s = N_("Show tips on start up");
gchar *s = N_("Previous");
gchar *s = N_("Next");
gchar *s = N_("Close");
gchar *s = N_("Preferences");
gchar *s = N_("Patches");
gchar *s = N_("Samples");
gchar *s = N_("Search Path");
gchar *s = N_("Browse");
gchar *s = N_("Browse");
gchar *s = N_("Browse");
gchar *s = N_("Colon separated list of directories to search for");
gchar *s = N_("Default Paths");
gchar *s = N_("Swami Tips");
gchar *s = N_("Splash image");
gchar *s = N_("Restore pane geometry");
gchar *s = N_("Startup Options");
gchar *s = N_("Confirm Quit");
gchar *s = N_("Always");
gchar *s = N_("Unsaved");
gchar *s = N_("Never");
gchar *s = N_("Save window geometry");
gchar *s = N_("Save Now");
gchar *s = N_("Exit Options");
gchar *s = N_("General");
gchar *s = N_("Bank");
gchar *s = N_("Preset");
gchar *s = N_("Temporary Audible");
gchar *s = N_("Text appended to stereo channel names");
gchar *s = N_("Left");
gchar *s = N_("Right");
gchar *s = N_("Max swap file waste");
gchar *s = N_("MB");
gchar *s = N_("Patches");
gchar *s = N_("Lower Octave");
gchar *s = N_("Upper Octave");
gchar *s = N_("Change");
gchar *s = N_("Change All");
gchar *s = N_("Piano");
gchar *s = N_("OK");
gchar *s = N_("Save Preferences");
gchar *s = N_("Cancel");
gchar *s = N_("Paste conflict");
gchar *s = N_("Source");
gchar *s = N_("Modify source item");
gchar *s = N_("Modify");
gchar *s = N_("Destination");
gchar *s = N_("Modify conflict item");
gchar *s = N_("Modify");
gchar *s = N_("For");
gchar *s = N_("Remember action for other items");
gchar *s = N_("This Conflict");
gchar *s = N_("This Type");
gchar *s = N_("All");
gchar *s = N_("Skip this paste item");
gchar *s = N_("Skip");
gchar *s = N_("Replace destination");
gchar *s = N_("Replace");
gchar *s = N_("Cancel entire paste operation");
gchar *s = N_("Cancel");
gchar *s = N_("Undo History");
gchar *s = N_("Backward in state history");
gchar *s = N_("Back");
gchar *s = N_("Forward in state history");
gchar *s = N_("Forward");
gchar *s = N_("Undo previous action");
gchar *s = N_("Undo");
gchar *s = N_("Redo next action");
gchar *s = N_("Redo");
gchar *s = N_("Jump to state of selected item");
gchar *s = N_("Jump");
gchar *s = N_("Close");
gchar *s = N_("Name:");
gchar *s = N_("Author:");
gchar *s = N_("Copyright:");
gchar *s = N_("Product:");
gchar *s = N_("Date:");
gchar *s = N_("Today");
gchar *s = N_("Version:");
gchar *s = N_("Sound Engine:");
gchar *s = N_("ROM:");
gchar *s = N_("Created With:");
gchar *s = N_("Edited With:");
gchar *s = N_("Comment");
gchar *s = N_("Chan");
gchar *s = N_("Preset");
gchar *s = N_("Bank");
gchar *s = N_("New");
gchar *s = N_("Delete");
gchar *s = N_("Source");
gchar *s = N_("Destination");
gchar *s = N_("Amount");
gchar *s = N_("Amount Source");
gchar *s = N_("Root Note");
gchar *s = N_("Exclusive Class");
gchar *s = N_("Fixed Note");
gchar *s = N_("Fixed Velocity");
gchar *s = N_("Set");
gchar *s = N_("Set");
gchar *s = N_("Set");
gchar *s = N_("Set");
gchar *s = N_("File");
gchar *s = N_("Edit");
gchar *s = N_("Plugins");
gchar *s = N_("Help");
gchar *s = N_("Swami Tips");
gchar *s = N_("About");
gchar *s = N_("Swami");
gchar *s = N_("Copyright (C) 1999-2006 Josh Green");
gchar *s = N_("Distributed under the GNU General Public License");
gchar *s = N_("translator-credits");
gchar *s = N_("Loop Finder");
gchar *s = N_("Sample");
gchar *s = N_("Length");
gchar *s = N_("<b>Results</b>");
gchar *s = N_("Min loop size");
gchar *s = N_("Size of analysis window (affects quality of results)");
gchar *s = N_("Minimum length of matching loops");
gchar *s = N_("Maximum loop suggestion results");
gchar *s = N_("Analysis window");
gchar *s = N_("Max results");
gchar *s = N_("Loop end");
gchar *s = N_("Loop start");
gchar *s = N_("Loop end search area position");
gchar *s = N_("Loop start search area position");
gchar *s = N_("Position");
gchar *s = N_("Loop start search area size");
gchar *s = N_("Loop end search area size");
gchar *s = N_("Search entire sample for loop start");
gchar *s = N_("Entire Sample");
gchar *s = N_("Search entire sample for loop end");
gchar *s = N_("Entire sample");
gchar *s = N_("Size");
gchar *s = N_("<b>Search area</b>");
gchar *s = N_("Revert to original loop values");
gchar *s = N_("0.00 secs");
gchar *s = N_("window1");
gchar *s = N_(" Name");
gchar *s = N_("Save current script");
gchar *s = N_("Create new script");
gchar *s = N_("Execute current script");
gchar *s = N_("Execute a line when ENTER is pressed");
gchar *s = N_("Shell mode");
gchar *s = N_("<b>Script Editor</b>");
gchar *s = N_("<b>Python Console</b>");
